module Ex04 where

import Data.Semigroup
import Data.Monoid
--import Control.Monad
import Control.Monad.State (State, get, put, evalState, execState, runState)
import Test.QuickCheck
import Data.Tuple(swap) -- This just swaps the vaues in a tuple, which is useful
-- for testing pop against the output of runState
import Priority
import Size

-- DEFINITIONS AND HELPER FUNCTIONS --

type NodeInfo = (Size, Priority)
data QueueTree a
  = Null
  | Leaf NodeInfo a
  | Node NodeInfo (QueueTree a) (QueueTree a)
  deriving (Show, Eq)

nodeInfo :: QueueTree a -> NodeInfo
nodeInfo Null = mempty
nodeInfo (Leaf i _) = i
nodeInfo (Node i _ _) = i

sizeOf :: QueueTree a -> Size
sizeOf = fst . nodeInfo

maxPrio :: QueueTree a -> Priority
maxPrio = snd . nodeInfo

-- checks whether the tree structure
-- is balanced (i.e. that the left subtree and the right
-- subtree don't ever differ too much in size)
balanced :: QueueTree a -> Bool
balanced (Node i l r) =
  let sl = unSize (sizeOf l) in
  let sr = unSize (sizeOf r) in
  abs (sl - sr) <= 1 && balanced l && balanced r
balanced _ = True


-- EXERCISE STARTS HERE --

-- Task 1a. Write a well-formedness predicate
--          wf for the `QueueTree` data type.

-- Hint: Both `Priority` and `Size` are semigroups/monoids.
-- This means that the type `NodeInfo` is also automatically
-- a monoid.

getInfo :: QueueTree a -> NodeInfo
getInfo t = go t mempty where
    go Null acc = acc
    go (Leaf j a) acc = j <> acc
    go (Node i l r) acc = (go l acc) <> (go r acc)

wf :: QueueTree a -> Bool
wf Null = True
wf (Leaf i a) = (fst i) == size 1 -- unsize (sizeOf (Leaf i a)) == 1
wf (Node i l r) = (i == nodeInfo l <> nodeInfo r) && wf l && wf r

-- This sort of thing is why I defined the getInfo function :)
wf_correct t = (nodeInfo t == getInfo t) == wf t

test = Node (size 3, priority 3) (Node (size 2, priority 3) (Leaf (size 1, priority 3) "b") (Leaf (size 1, priority 2) "a")) (Leaf (size 1, priority 3) "c")

-- Task 1b. Write smart constructors `leaf` and `node`
--          for the `QueueTree` data type which maintain
--          the well-formedness invariant. I.e. given
--          well-formed inputs, the smart constructors
--          should give well-formed outputs.


leaf :: Priority -> a -> QueueTree a
leaf p a = Leaf (size 1, p) a

leaf_correct p a = wf $ leaf p a

node :: QueueTree a -> QueueTree a -> QueueTree a
node l r = Node (nodeInfo l <> nodeInfo r) l r

node_correct l r = wf $ node l r

-- Task 2a. Implement the usual priority queue functions
--          for the type `QueueTree`. These are
--          pop - Remove the element from the queue that has the
--               highest priority. Return the modified queue,
--               along with the removed element (if any).
--          insert - add an element to the queue with the given priority.

pop :: QueueTree a -> (QueueTree a, Maybe a)
pop Null = (Null, Nothing)
pop (Leaf i a) = (Null, Just a)
pop (Node i l r) | maxPrio l == (snd i) = let (l', a) = pop l
                                          in (node l' r, a)
                 | otherwise            = let (r', a) = pop r
                                          in (node l r', a)



insert :: Priority -> a -> QueueTree a -> QueueTree a
insert p a Null = leaf p a
insert p a (Leaf i b) = node (leaf p a) (Leaf i b)
insert p a (Node i l r) = if sl <= sr then node (insert p a l) r else node l (insert p a r) where
    sl = sizeOf l
    sr = sizeOf r

insert_correct :: Priority -> a -> QueueTree a -> Bool
insert_correct p a t = balanced (insert p a t) == balanced t  

-- Task 2b. Implement a function `fromList` that converts a
--          list of `(Priority, x)` pairs into a well-formed
--          and balanced `QueueTree x` structure.

fromList :: [(Priority, a)] -> QueueTree a
fromList xs = foldr (uncurry insert) Null xs


instance (Arbitrary a) => Arbitrary (QueueTree a) where
    arbitrary = fromList <$> arbitrary

-- Since inserting preserves being balanced, this only produces balanced trees
-- which we can check with the property below. This constirutes a check for fromList
-- as well since our arbitrary instance just calls fromList on arbitrary lists
balanced_check :: QueueTree a -> Bool
balanced_check t = balanced t 

-- Hint: you can use `fromList` to implement an `Arbitrary`
-- instance for `QueueTree`, allowing you to test your work.


-- Task 3. Implement stateful versions of the pop and insert
--         operations above using the `State` type in Haskell's
--         standard mtl library.
--         Implement a `peek` operation which just returns the
--         highest-priority element without changing the
--         state of the queue.
--         Do not use the `state` function in your final
--         implementations!


{-
Recall:

(>>=) :: State s a -> (a -> State s b) -> State s b
f >>= rest = \origState ->
    let (newState, retVal) = f origState in
    rest retVal newState

State s a = s -> (s, a)

get :: State s s

return :: a -> State s a
return v = \origState -> (origState, v)

(>>) :: State s a -> State s b -> State s b
f >> sb = f >>= \d -> sb
-}

-- put :: State s ()
pop' :: State (QueueTree a) (Maybe a) -- QueueTree a -> (Queuetree a, Maybe a)
pop' = 
    get >>= \t ->
    let (t', a) = pop t
    in put t' >> -- same thing as: in put t' >>= \() ->
    return a

pop'_correct :: (Eq a) => QueueTree a -> Bool
pop'_correct t = swap (runState pop' t) == pop t 

pop_do :: State (QueueTree a) (Maybe a)
pop_do = do
    t <- get
    let (t', a) = pop t
    put t'
    return a

pop_do_correct :: Eq a => QueueTree a -> Bool
pop_do_correct t = runState pop' t == runState pop_do t

insert' :: Priority -> a -> State (QueueTree a) ()
insert' p a =
    get >>= \t ->
    let t' = insert p a t 
    in put t'

insert'_correct :: Eq a => Priority -> a -> QueueTree a -> Bool
insert'_correct p a t = execState (insert' p a) t == insert p a t

peek' :: State (QueueTree a) (Maybe a)
peek' = do
    t <- get
    ret <- pop'
    put t
    return ret

peek'' :: State (QueueTree a) (Maybe a)
peek'' = do
    t <- get
    let (t', a) = pop t
    return a

-------------------------------------------------------------------

{-
I was asked in class to provide some examples of implementations of pop'
and insert' which don't call the pop and insert functions.

To be clear, this is not a good idea; you've already implemented the logic
for these functions and having access to things like get and put doesn't
help you implement these much. In fact, writing it this way is a bit worse
because we're still just wrapping the functions in binds and returns, it's 
just that we're doing so in a much more verbose way.

Nevertheless, here are some implementations.
-}

pop'_verbose :: State (QueueTree a) (Maybe a)
pop'_verbose = 
    get >>= \t ->
    go t where
        go Null = return Nothing
        go (Leaf i a) = put Null >> return (Just a)
        go (Node i l r) | maxPrio l == (snd i) = put l >>
                                                 pop'_verbose >>= \b ->
                                                 get >>= \l' ->
                                                 put (node l' r) >>
                                                 return b
                        | otherwise            = put r >>
                                                 pop'_verbose >>= \c ->
                                                 get >>= \r' ->
                                                 put (node l r') >>
                                                 return c

insert'_verbose :: Priority -> a -> State (QueueTree a) ()
insert'_verbose p a =
    get >>= \t ->
    go t where
        go Null = put $ leaf p a
        go (Leaf i b) = put $ node (leaf p a) (Leaf i b)
        go (Node i l r) | sl <= sr  = put l >>
                                      insert'_verbose p a >>
                                      get >>= \l' ->
                                      put $ node l' r
                        | otherwise = put r >>= \_ -> -- same as >> above
                                      insert'_verbose p a >>
                                      get >>= \r' ->
                                      put $ node l r' where
                                        sl = sizeOf l
                                        sr = sizeOf r

{- 
These properties below pass with quickCheck, so these functions do indeed do the same thing as
the simpler implementations of pop' and insert' (or at least we can be pretty sure they do :) )
-}
pop'_verbose_correct :: Eq a => QueueTree a -> Bool
pop'_verbose_correct t = runState pop' t == runState pop'_verbose t

insert'_verbose_correct :: Eq a => Priority -> a -> QueueTree a -> Bool
insert'_verbose_correct p a t = runState (insert' p a) t == runState (insert'_verbose p a) t
-- END OF EXERCISE --

-- You can use the following three examples to test your
-- implementations of pop' and insert', and to practice
-- reading `State`-ful functions.

-- Returns the highest priority currently in the `QueueTree`
-- without changing the state.
getMaxPrio' :: State (QueueTree a) Priority
getMaxPrio' =
  get >>= \q ->
  return (maxPrio q)

-- Removes the element with the second-highest priority
-- in the queue.
dip' :: State (QueueTree a) () -- QueueTree a -> (QueueTree a, ())
dip' =
  getMaxPrio' >>= \p ->
  pop'        >>= \h1 ->
  pop'        >>= \h2 ->
  case h1 of
    Nothing -> return ()
    Just h1 -> insert' p h1

-- a `State`-free version of dip
dip :: QueueTree Char -> QueueTree Char
dip = evalState $
  dip' >>= \() ->
  get
